# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.pipeline import ClientRawResponse
from msrest.exceptions import HttpOperationError

from .. import models


class BulkRegistryOperations(object):
    """BulkRegistryOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    :ivar api_version: Version of the Api. Constant value: "2021-04-12".
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config
        self.api_version = "2021-04-12"

    def update_registry(self, devices, custom_headers=None, raw=False, **operation_config):
        """Creates, updates, or deletes the identities of multiple devices from
        the IoT Hub identity registry. A device identity can be specified only
        once in the list. Different operations (create, update, delete) on
        different devices are allowed. A maximum of 100 devices can be
        specified per invocation. For large scale operations, use the import
        feature using blob storage
        (https://docs.microsoft.com/azure/iot-hub/iot-hub-devguide-identity-registry#import-and-export-device-identities).

        :param devices: The registry operations to perform.
        :type devices: list[~protocol.models.ExportImportDevice]
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: BulkRegistryOperationResult or ClientRawResponse if raw=true
        :rtype: ~protocol.models.BulkRegistryOperationResult or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.update_registry.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(devices, "[ExportImportDevice]")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200, 400]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("BulkRegistryOperationResult", response)
        if response.status_code == 400:
            deserialized = self._deserialize("BulkRegistryOperationResult", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    update_registry.metadata = {"url": "/devices"}



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ConfigurationQueriesTestResponse(Model):
    """ConfigurationQueriesTestResponse.

    :param target_condition_error: The errors from running the target
     condition query.
    :type target_condition_error: str
    :param custom_metric_query_errors: The errors from running the custom
     metric query.
    :type custom_metric_query_errors: dict[str, str]
    """

    _attribute_map = {
        "target_condition_error": {"key": "targetConditionError", "type": "str"},
        "custom_metric_query_errors": {"key": "customMetricQueryErrors", "type": "{str}"},
    }

    def __init__(
        self, *, target_condition_error: str = None, custom_metric_query_errors=None, **kwargs
    ) -> None:
        super(ConfigurationQueriesTestResponse, self).__init__(**kwargs)
        self.target_condition_error = target_condition_error
        self.custom_metric_query_errors = custom_metric_query_errors



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ConfigurationContent(Model):
    """The configuration content for devices or modules on edge devices.

    :param device_content: The device configuration content.
    :type device_content: dict[str, object]
    :param modules_content: The modules configuration content.
    :type modules_content: dict[str, dict[str, object]]
    :param module_content: The module configuration content.
    :type module_content: dict[str, object]
    """

    _attribute_map = {
        "device_content": {"key": "deviceContent", "type": "{object}"},
        "modules_content": {"key": "modulesContent", "type": "{{object}}"},
        "module_content": {"key": "moduleContent", "type": "{object}"},
    }

    def __init__(
        self, *, device_content=None, modules_content=None, module_content=None, **kwargs
    ) -> None:
        super(ConfigurationContent, self).__init__(**kwargs)
        self.device_content = device_content
        self.modules_content = modules_content
        self.module_content = module_content



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class JobProperties(Model):
    """JobProperties.

    :param job_id: The unique identifier of the job.
    :type job_id: str
    :param start_time_utc: System generated. Ignored at creation. The start
     date and time of the job in UTC.
    :type start_time_utc: datetime
    :param end_time_utc: System generated. Ignored at creation. The end date
     and time of the job in UTC.
    :type end_time_utc: datetime
    :param type: The job type. Possible values include: 'unknown', 'export',
     'import', 'backup', 'readDeviceProperties', 'writeDeviceProperties',
     'updateDeviceConfiguration', 'rebootDevice', 'factoryResetDevice',
     'firmwareUpdate', 'scheduleDeviceMethod', 'scheduleUpdateTwin',
     'restoreFromBackup', 'failoverDataCopy'
    :type type: str or ~protocol.models.enum
    :param status: System generated. Ignored at creation. The status of the
     job. Possible values include: 'unknown', 'enqueued', 'running',
     'completed', 'failed', 'cancelled', 'scheduled', 'queued'
    :type status: str or ~protocol.models.enum
    :param progress: System generated. Ignored at creation. The percentage of
     job completion.
    :type progress: int
    :param input_blob_container_uri: The URI containing SAS token to a blob
     container that contains registry data to sync.
    :type input_blob_container_uri: str
    :param input_blob_name: The blob name to use when importing from the input
     blob container.
    :type input_blob_name: str
    :param output_blob_container_uri: The SAS token to access the blob
     container. This is used to output the status and results of the job.
    :type output_blob_container_uri: str
    :param output_blob_name: The blob name that will be created in the output
     blob container. This blob will contain the exported device registry
     information for the IoT Hub.
    :type output_blob_name: str
    :param exclude_keys_in_export: Optional for export jobs; ignored for other
     jobs. If not specified, the service defaults to false. If false,
     authorization keys are included in export output. Keys are exported as
     null otherwise.
    :type exclude_keys_in_export: bool
    :param storage_authentication_type: The authentication type used for
     connecting to the storage account. Possible values include: 'keyBased',
     'identityBased'
    :type storage_authentication_type: str or ~protocol.models.enum
    :param identity:
    :type identity: ~protocol.models.ManagedIdentity
    :param failure_reason: System genereated.  Ignored at creation. The reason
     for failure, if a failure occurred.
    :type failure_reason: str
    :param include_configurations: Defaults to false. If true, then
     configurations are included in the data export/import.
    :type include_configurations: bool
    :param configurations_blob_name: Defaults to configurations.txt. Specifies
     the name of the blob to use when exporting/importing configurations.
    :type configurations_blob_name: str
    """

    _attribute_map = {
        "job_id": {"key": "jobId", "type": "str"},
        "start_time_utc": {"key": "startTimeUtc", "type": "iso-8601"},
        "end_time_utc": {"key": "endTimeUtc", "type": "iso-8601"},
        "type": {"key": "type", "type": "str"},
        "status": {"key": "status", "type": "str"},
        "progress": {"key": "progress", "type": "int"},
        "input_blob_container_uri": {"key": "inputBlobContainerUri", "type": "str"},
        "input_blob_name": {"key": "inputBlobName", "type": "str"},
        "output_blob_container_uri": {"key": "outputBlobContainerUri", "type": "str"},
        "output_blob_name": {"key": "outputBlobName", "type": "str"},
        "exclude_keys_in_export": {"key": "excludeKeysInExport", "type": "bool"},
        "storage_authentication_type": {"key": "storageAuthenticationType", "type": "str"},
        "identity": {"key": "identity", "type": "ManagedIdentity"},
        "failure_reason": {"key": "failureReason", "type": "str"},
        "include_configurations": {"key": "includeConfigurations", "type": "bool"},
        "configurations_blob_name": {"key": "configurationsBlobName", "type": "str"},
    }

    def __init__(self, **kwargs):
        super(JobProperties, self).__init__(**kwargs)
        self.job_id = kwargs.get("job_id", None)
        self.start_time_utc = kwargs.get("start_time_utc", None)
        self.end_time_utc = kwargs.get("end_time_utc", None)
        self.type = kwargs.get("type", None)
        self.status = kwargs.get("status", None)
        self.progress = kwargs.get("progress", None)
        self.input_blob_container_uri = kwargs.get("input_blob_container_uri", None)
        self.input_blob_name = kwargs.get("input_blob_name", None)
        self.output_blob_container_uri = kwargs.get("output_blob_container_uri", None)
        self.output_blob_name = kwargs.get("output_blob_name", None)
        self.exclude_keys_in_export = kwargs.get("exclude_keys_in_export", None)
        self.storage_authentication_type = kwargs.get("storage_authentication_type", None)
        self.identity = kwargs.get("identity", None)
        self.failure_reason = kwargs.get("failure_reason", None)
        self.include_configurations = kwargs.get("include_configurations", None)
        self.configurations_blob_name = kwargs.get("configurations_blob_name", None)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from .auth import ConnectionStringAuthentication, AzureIdentityCredentialAdapter
from .protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs as protocol_client
from .protocol.models import Configuration, ConfigurationContent, ConfigurationQueriesTestInput


class IoTHubConfigurationManager(object):
    """A class to provide convenience APIs for IoTHub Configuration Manager operations,
    based on top of the auto generated IotHub REST APIs
    """

    def __init__(self, connection_string=None, host=None, auth=None):
        """Initializer for a Configuration Manager Service client.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub if we are using connection_str authentication. Default value: None
        :param str host: The Azure service url if we are using token credential authentication.
            Default value: None
        :param str auth: The Azure authentication object if we are using token credential authentication.
            Default value: None

        :returns: Instance of the IoTHubConfigurationManager object.
        :rtype: :class:`azure.iot.hub.IoTHubConfigurationManager`
        """
        if connection_string is not None:
            self.auth = ConnectionStringAuthentication(connection_string)
            self.protocol = protocol_client(self.auth, "https://" + self.auth["HostName"])
        else:
            self.auth = auth
            self.protocol = protocol_client(self.auth, "https://" + host)

    @classmethod
    def from_connection_string(cls, connection_string):
        """Classmethod initializer for a IoTHubConfigurationManager Service client.
        Creates IoTHubConfigurationManager class from connection string.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub.

        :rtype: :class:`azure.iot.hub.IoTHubConfigurationManager`
        """
        return cls(connection_string=connection_string)

    @classmethod
    def from_token_credential(cls, url, token_credential):
        """Classmethod initializer for a IoTHubConfigurationManager Service client.
        Creates IoTHubConfigurationManager class from host name url and Azure token credential.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str url: The Azure service url (host name).
        :param str token_credential: The Azure token credential object.

        :rtype: :class:`azure.iot.hub.IoTHubConfigurationManager`
        """
        host = url
        auth = AzureIdentityCredentialAdapter(token_credential)
        return cls(host=host, auth=auth)

    def get_configuration(self, configuration_id):
        """Retrieves the IoTHub configuration for a particular device.

        :param str configuration_id: The id of the configuration.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Configuration object.
        """
        return self.protocol.configuration.get(configuration_id)

    def create_configuration(self, configuration):
        """Creates a configuration for devices or modules of an IoTHub.

        :param str configuration_id: The id of the configuration.
        :param Configuration configuration: The configuration to create.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Configuration object containing the created configuration.
        """
        return self.protocol.configuration.create_or_update(configuration.id, configuration)

    def update_configuration(self, configuration, etag):
        """Updates a configuration for devices or modules of an IoTHub.
           Note: that configuration Id and Content cannot be updated by the user.

        :param str configuration_id: The id of the configuration.
        :param Configuration configuration: The configuration contains the updated configuration.
        :param str etag: The etag (if_match) value to use for the update operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Configuration object containing the updated configuration.
        """
        return self.protocol.configuration.create_or_update(configuration.id, configuration, etag)

    def delete_configuration(self, configuration_id, etag=None):
        """Deletes a configuration from an IoTHub.

        :param str configuration_id: The id of the configuration.
        :param Configuration configuration: The configuration to create.
        :param str etag: The etag (if_match) value to use for the delete operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Configuration object containing the updated configuration.
        """
        if etag is None:
            etag = "*"

        return self.protocol.configuration.delete(configuration_id, etag)

    def get_configurations(self, max_count=None):
        """Retrieves multiple configurations for device and modules of an IoTHub.
           Returns the specified number of configurations. Pagination is not supported.

        :param int max_count: The maximum number of configurations requested.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The list[Configuration] object.
        """
        return self.protocol.configuration.get_configurations(max_count)

    def test_configuration_queries(self, configuration_queries_test_input):
        """Validates the target condition query and custom metric queries for a
           configuration.

        :param ConfigurationQueriesTestInput configuration_queries_test_input: The queries test input.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The ConfigurationQueriesTestResponse object.
        """
        return self.protocol.configuration.test_queries(configuration_queries_test_input)

    def apply_configuration_on_edge_device(self, device_id, configuration_content):
        """Applies the provided configuration content to the specified edge
           device. Modules content is mandantory.

        :param ConfigurationContent configuration_content: The name (Id) of the edge device.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: An object.
        """
        return self.protocol.configuration.apply_on_edge_device(device_id, configuration_content)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import os
import sys
import base64
import time
import hashlib
import hmac
from uuid import uuid4
import six.moves.urllib as urllib

try:
    from urllib import quote, quote_plus, urlencode  # Py2
except Exception:
    from urllib.parse import quote, quote_plus, urlencode

import uamqp

default_sas_expiry = 30


class IoTHubAmqpClient:
    def _generate_auth_token(self, uri, sas_name, sas_value):
        sas = base64.b64decode(sas_value)
        expiry = str(int(time.time() + default_sas_expiry))
        string_to_sign = (uri + "\n" + expiry).encode("utf-8")
        signed_hmac_sha256 = hmac.HMAC(sas, string_to_sign, hashlib.sha256)
        signature = urllib.parse.quote(base64.b64encode(signed_hmac_sha256.digest()))
        return "SharedAccessSignature sr={}&sig={}&se={}&skn={}".format(
            uri, signature, expiry, sas_name
        )

    def _build_amqp_endpoint(self, hostname, shared_access_key_name, shared_access_key):
        hub_name = hostname.split(".")[0]
        endpoint = "{}@sas.root.{}".format(shared_access_key_name, hub_name)
        endpoint = quote_plus(endpoint)
        sas_token = self._generate_auth_token(
            hostname, shared_access_key_name, shared_access_key + "="
        )
        endpoint = endpoint + ":{}@{}".format(quote_plus(sas_token), hostname)
        return endpoint

    def __init__(self, hostname, shared_access_key_name, shared_access_key):
        self.endpoint = self._build_amqp_endpoint(
            hostname, shared_access_key_name, shared_access_key
        )
        operation = "/messages/devicebound"
        target = "amqps://" + self.endpoint + operation
        self.amqp_client = uamqp.SendClient(target)

    def disconnect_sync(self):
        """
        Disconnect the Amqp client.
        """
        if self.amqp_client:
            self.amqp_client.close()
            self.amqp_client = None

    def send_message_to_device(self, device_id, message, app_props):
        """Send a message to the specified deivce.

        :param str device_id: The name (Id) of the device.
        :param str message: The message that is to be delivered to the device.
        :param dict app_props: Application and system properties for the message

        :raises: Exception if the Send command is not able to send the message
        """
        msg_content = message
        msg_props = uamqp.message.MessageProperties()
        msg_props.message_id = str(uuid4())
        msg_props.to = "/devices/{}/messages/devicebound".format(device_id)

        app_properties = {}

        # loop through all properties and pull out the custom
        # properties
        for prop_key, prop_value in app_props.items():
            if prop_key == "contentType":
                msg_props.content_type = prop_value
            elif prop_key == "contentEncoding":
                msg_props.content_encoding = prop_value
            elif prop_key == "correlationId":
                msg_props.correlation_id = prop_value
            elif prop_key == "expiryTimeUtc":
                msg_props.absolute_expiry_time = prop_value
            elif prop_key == "messageId":
                msg_props.message_id = prop_value
            else:
                app_properties[prop_key] = prop_value

        message = uamqp.Message(
            msg_content, properties=msg_props, application_properties=app_properties
        )
        self.amqp_client.queue_message(message)
        results = self.amqp_client.send_all_messages(close_on_done=False)
        if uamqp.constants.MessageState.SendFailed in results:
            raise Exception("C2D message send failure")



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from .configuration_operations import ConfigurationOperations
from .statistics_operations import StatisticsOperations
from .devices_operations import DevicesOperations
from .bulk_registry_operations import BulkRegistryOperations
from .query_operations import QueryOperations
from .jobs_operations import JobsOperations
from .cloud_to_device_messages_operations import CloudToDeviceMessagesOperations
from .modules_operations import ModulesOperations
from .digital_twin_operations import DigitalTwinOperations

__all__ = [
    "ConfigurationOperations",
    "StatisticsOperations",
    "DevicesOperations",
    "BulkRegistryOperations",
    "QueryOperations",
    "JobsOperations",
    "CloudToDeviceMessagesOperations",
    "ModulesOperations",
    "DigitalTwinOperations",
]



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.pipeline import ClientRawResponse
from msrest.exceptions import HttpOperationError

from .. import models


class StatisticsOperations(object):
    """StatisticsOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    :ivar api_version: Version of the Api. Constant value: "2021-04-12".
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config
        self.api_version = "2021-04-12"

    def get_device_statistics(self, custom_headers=None, raw=False, **operation_config):
        """Gets device statistics of the IoT Hub identity registry, such as total
        device count.

        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: RegistryStatistics or ClientRawResponse if raw=true
        :rtype: ~protocol.models.RegistryStatistics or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_device_statistics.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("RegistryStatistics", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_device_statistics.metadata = {"url": "/statistics/devices"}

    def get_service_statistics(self, custom_headers=None, raw=False, **operation_config):
        """Gets service statistics of the IoT Hub identity registry, such as
        connected device count.

        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: ServiceStatistics or ClientRawResponse if raw=true
        :rtype: ~protocol.models.ServiceStatistics or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_service_statistics.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("ServiceStatistics", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_service_statistics.metadata = {"url": "/statistics/service"}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from .iothub_amqp_client import IoTHubAmqpClient as iothub_amqp_client
from .auth import ConnectionStringAuthentication, AzureIdentityCredentialAdapter
from .protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs as protocol_client
from .protocol.models import (
    Device,
    Module,
    SymmetricKey,
    X509Thumbprint,
    AuthenticationMechanism,
    ServiceStatistics,
    RegistryStatistics,
    QuerySpecification,
    Twin,
    CloudToDeviceMethod,
    CloudToDeviceMethodResult,
    DeviceCapabilities,
)


def _ensure_quoted(etag):
    if not isinstance(etag, str) or (len(etag) > 1 and etag[0] == '"' and etag[-1] == '"'):
        return etag
    return '"' + etag + '"'


class QueryResult(object):
    """The query result.
    :param type: The query result type. Possible values include: 'unknown', 'twin', 'deviceJob', 'jobResponse', 'raw', 'enrollment', 'enrollmentGroup', 'deviceRegistration'
    :type type: str or ~protocol.models.enum
    :param items: The query result items, as a collection.
    :type items: list[object]
    :param continuation_token: Request continuation token.
    :type continuation_token: str
    """

    _attribute_map = {
        "type": {"key": "type", "type": "str"},
        "items": {"key": "items", "type": "[object]"},
        "continuation_token": {"key": "continuationToken", "type": "str"},
    }

    def __init__(self, **kwargs):
        super(QueryResult, self).__init__(**kwargs)
        self.type = kwargs.get("type", None)
        self.items = kwargs.get("items", None)
        self.continuation_token = kwargs.get("continuation_token", None)


class IoTHubRegistryManager(object):
    """A class to provide convenience APIs for IoTHub Registry Manager operations,
    based on top of the auto generated IotHub REST APIs
    """

    def __init__(self, connection_string=None, host=None, auth=None):
        """Initializer for a Registry Manager Service client.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub if we are using connection_str authentication. Default value: None
        :param str host: The Azure service url if we are using token credential authentication.
            Default value: None
        :param str auth: The Azure authentication object if we are using token credential authentication.
            Default value: None

        :returns: Instance of the IoTHubRegistryManager object.
        :rtype: :class:`azure.iot.hub.IoTHubRegistryManager`
        """
        if connection_string is not None:
            self.auth = ConnectionStringAuthentication(connection_string)
            self.protocol = protocol_client(self.auth, "https://" + self.auth["HostName"])
            self.amqp_svc_client = iothub_amqp_client(
                self.auth["HostName"],
                self.auth["SharedAccessKeyName"],
                self.auth["SharedAccessKey"],
            )
        else:
            self.auth = auth
            self.protocol = protocol_client(self.auth, "https://" + host)
            self.amqp_svc_client = None

    @classmethod
    def from_connection_string(cls, connection_string):
        """Classmethod initializer for a Registry Manager Service client.
        Creates Registry Manager class from connection string.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub.

        :rtype: :class:`azure.iot.hub.IoTHubRegistryManager`
        """
        return cls(connection_string=connection_string)

    @classmethod
    def from_token_credential(cls, url, token_credential):
        """Classmethod initializer for a Registry Manager Service client.
        Creates Registry Manager class from host name url and Azure token credential.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str url: The Azure service url (host name).
        :param str token_credential: The Azure token credential object.

        :rtype: :class:`azure.iot.hub.IoTHubRegistryManager`
        """
        host = url
        auth = AzureIdentityCredentialAdapter(token_credential)
        return cls(host=host, auth=auth)

    def __del__(self):
        """
        Deinitializer for a Registry Manager Service client.
        """
        if self.amqp_svc_client:
            self.amqp_svc_client.disconnect_sync()

    def create_device_with_sas(
        self,
        device_id,
        primary_key,
        secondary_key,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Creates a device identity on IoTHub using SAS authentication.

        :param str device_id: The name (Id) of the device.
        :param str primary_key: Primary authentication key.
        :param str secondary_key: Secondary authentication key.
        :param str status: Initital state of the created device.
            (Possible values: "enabled" or "disabled")
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Device object containing the created device.
        """
        symmetric_key = SymmetricKey(primary_key=primary_key, secondary_key=secondary_key)

        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(type="sas", symmetric_key=symmetric_key),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        return self.protocol.devices.create_or_update_identity(device_id, device)

    def create_device_with_x509(
        self,
        device_id,
        primary_thumbprint,
        secondary_thumbprint,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Creates a device identity on IoTHub using X509 authentication.

        :param str device_id: The name (Id) of the device.
        :param str primary_thumbprint: Primary X509 thumbprint.
        :param str secondary_thumbprint: Secondary X509 thumbprint.
        :param str status: Initital state of the created device.
            (Possible values: "enabled" or "disabled")
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Device object containing the created device.
        """
        x509_thumbprint = X509Thumbprint(
            primary_thumbprint=primary_thumbprint, secondary_thumbprint=secondary_thumbprint
        )

        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(
                type="selfSigned", x509_thumbprint=x509_thumbprint
            ),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        return self.protocol.devices.create_or_update_identity(device_id, device)

    def create_device_with_certificate_authority(
        self,
        device_id,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Creates a device identity on IoTHub using certificate authority.

        :param str device_id: The name (Id) of the device.
        :param str status: Initial state of the created device.
            (Possible values: "enabled" or "disabled").
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Device object containing the created device.
        """
        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(type="certificateAuthority"),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        return self.protocol.devices.create_or_update_identity(device_id, device)

    def update_device_with_sas(
        self,
        device_id,
        etag,
        primary_key,
        secondary_key,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Updates a device identity on IoTHub using SAS authentication.

        :param str device_id: The name (Id) of the device.
        :param str etag: The etag (if_match) value to use for the update operation.
        :param str primary_key: Primary authentication key.
        :param str secondary_key: Secondary authentication key.
        :param str status: Initital state of the created device.
            (Possible values: "enabled" or "disabled").
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Device object containing the created device.
        """
        symmetric_key = SymmetricKey(primary_key=primary_key, secondary_key=secondary_key)

        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(type="sas", symmetric_key=symmetric_key),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.devices.create_or_update_identity(
            device_id, device, _ensure_quoted(etag)
        )

    def update_device_with_x509(
        self,
        device_id,
        etag,
        primary_thumbprint,
        secondary_thumbprint,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Updates a device identity on IoTHub using X509 authentication.

        :param str device_id: The name (Id) of the device.
        :param str etag: The etag (if_match) value to use for the update operation.
        :param str primary_thumbprint: Primary X509 thumbprint.
        :param str secondary_thumbprint: Secondary X509 thumbprint.
        :param str status: Initital state of the created device.
            (Possible values: "enabled" or "disabled").
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Device object containing the created device.
        """
        x509_thumbprint = X509Thumbprint(
            primary_thumbprint=primary_thumbprint, secondary_thumbprint=secondary_thumbprint
        )

        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(
                type="selfSigned", x509_thumbprint=x509_thumbprint
            ),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.devices.create_or_update_identity(
            device_id, device, _ensure_quoted(etag)
        )

    def update_device_with_certificate_authority(
        self,
        device_id,
        etag,
        status,
        iot_edge=False,
        status_reason=None,
        device_scope=None,
        parent_scopes=None,
    ):
        """Updates a device identity on IoTHub using certificate authority.

        :param str device_id: The name (Id) of the device.
        :param str etag: The etag (if_match) value to use for the update operation.
        :param str status: Initital state of the created device.
            (Possible values: "enabled" or "disabled").
        :param bool iot_edge: Whether or not the created device is an IoT Edge device. Default value: False
        :param str status_reason: The reason for the device identity status. Default value: None
        :param str device_scope: The scope of the device. Default value: None
            Auto generated and immutable for edge devices and modifiable in leaf devices to create child/parent relationship.
            For leaf devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
        :param Union[list[str], str] parent_scopes: The scopes of the upper level edge devices if applicable. Default value: None
            For edge devices, the value to set a parent edge device can be retrieved from the parent edge device's device_scope property.
            For leaf devices, this could be set to the same value as device_scope or left for the service to copy over.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Device object containing the created device.
        """
        if isinstance(parent_scopes, str):
            parent_scopes = [parent_scopes]

        kwargs = {
            "device_id": device_id,
            "status": status,
            "authentication": AuthenticationMechanism(type="certificateAuthority"),
            "capabilities": DeviceCapabilities(iot_edge=iot_edge),
            "status_reason": status_reason,
            "device_scope": device_scope,
            "parent_scopes": parent_scopes,
        }
        device = Device(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.devices.create_or_update_identity(
            device_id, device, _ensure_quoted(etag)
        )

    def get_device(self, device_id):
        """Retrieves a device identity from IoTHub.

        :param str device_id: The name (Id) of the device.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Device object containing the requested device.
        """
        return self.protocol.devices.get_identity(device_id)

    def delete_device(self, device_id, etag=None):
        """Deletes a device identity from IoTHub.

        :param str device_id: The name (Id) of the device.
        :param str etag: The etag (if_match) value to use for the delete operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: None.
        """
        if etag is None:
            etag = "*"

        self.protocol.devices.delete_identity(device_id, _ensure_quoted(etag))

    def create_module_with_sas(self, device_id, module_id, managed_by, primary_key, secondary_key):
        """Creates a module identity for a device on IoTHub using SAS authentication.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).
        :param str primary_key: Primary authentication key.
        :param str secondary_key: Secondary authentication key.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Module object containing the created module.
        """
        symmetric_key = SymmetricKey(primary_key=primary_key, secondary_key=secondary_key)

        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "authentication": AuthenticationMechanism(type="sas", symmetric_key=symmetric_key),
        }
        module = Module(**kwargs)

        return self.protocol.modules.create_or_update_identity(device_id, module_id, module)

    def create_module_with_x509(
        self, device_id, module_id, managed_by, primary_thumbprint, secondary_thumbprint
    ):
        """Creates a module identity for a device on IoTHub using X509 authentication.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).
        :param str primary_thumbprint: Primary X509 thumbprint.
        :param str secondary_thumbprint: Secondary X509 thumbprint.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Module object containing the created module.
        """
        x509_thumbprint = X509Thumbprint(
            primary_thumbprint=primary_thumbprint, secondary_thumbprint=secondary_thumbprint
        )

        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "authentication": AuthenticationMechanism(
                type="selfSigned", x509_thumbprint=x509_thumbprint
            ),
        }
        module = Module(**kwargs)

        return self.protocol.modules.create_or_update_identity(device_id, module_id, module)

    def create_module_with_certificate_authority(self, device_id, module_id, managed_by):
        """Creates a module identity for a device on IoTHub using certificate authority.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Module object containing the created module.
        """
        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "authentication": AuthenticationMechanism(type="certificateAuthority"),
        }
        module = Module(**kwargs)

        return self.protocol.modules.create_or_update_identity(device_id, module_id, module)

    def update_module_with_sas(
        self, device_id, module_id, managed_by, etag, primary_key, secondary_key
    ):
        """Updates a module identity for a device on IoTHub using SAS authentication.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).
        :param str etag: The etag (if_match) value to use for the update operation.
        :param str primary_key: Primary authentication key.
        :param str secondary_key: Secondary authentication key.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Module object containing the created module.
        """
        symmetric_key = SymmetricKey(primary_key=primary_key, secondary_key=secondary_key)

        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "authentication": AuthenticationMechanism(type="sas", symmetric_key=symmetric_key),
        }
        module = Module(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.modules.create_or_update_identity(
            device_id, module_id, module, _ensure_quoted(etag)
        )

    def update_module_with_x509(
        self, device_id, module_id, managed_by, etag, primary_thumbprint, secondary_thumbprint
    ):
        """Updates a module identity for a device on IoTHub using X509 authentication.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).
        :param str etag: The etag (if_match) value to use for the update operation.
        :param str primary_thumbprint: Primary X509 thumbprint.
        :param str secondary_thumbprint: Secondary X509 thumbprint.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Module object containing the created module.
        """
        x509_thumbprint = X509Thumbprint(
            primary_thumbprint=primary_thumbprint, secondary_thumbprint=secondary_thumbprint
        )

        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "authentication": AuthenticationMechanism(
                type="selfSigned", x509_thumbprint=x509_thumbprint
            ),
        }
        module = Module(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.modules.create_or_update_identity(
            device_id, module_id, module, _ensure_quoted(etag)
        )

    def update_module_with_certificate_authority(self, device_id, module_id, managed_by, etag):
        """Updates a module identity for a device on IoTHub using certificate authority.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str managed_by: The name of the manager device (edge).
        :param str etag: The etag (if_match) value to use for the update operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The updated Module object containing the created module.
        """
        kwargs = {
            "device_id": device_id,
            "module_id": module_id,
            "managed_by": managed_by,
            "etag": etag,
            "authentication": AuthenticationMechanism(type="certificateAuthority"),
        }
        module = Module(**kwargs)

        if etag is None:
            etag = "*"

        return self.protocol.modules.create_or_update_identity(
            device_id, module_id, module, _ensure_quoted(etag)
        )

    def get_module(self, device_id, module_id):
        """Retrieves a module identity for a device from IoTHub.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Module object containing the requested module.
        """
        return self.protocol.modules.get_identity(device_id, module_id)

    def get_modules(self, device_id):
        """Retrieves all module identities on a device.

        :param str device_id: The name (Id) of the device.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The list[Module] containing all the modules on the device.
        """
        return self.protocol.modules.get_modules_on_device(device_id)

    def delete_module(self, device_id, module_id, etag=None):
        """Deletes a module identity for a device from IoTHub.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param str etag: The etag (if_match) value to use for the delete operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: None.
        """
        if etag is None:
            etag = "*"

        self.protocol.modules.delete_identity(device_id, module_id, _ensure_quoted(etag))

    def get_service_statistics(self):
        """Retrieves the IoTHub service statistics.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The ServiceStatistics object.
        """
        return self.protocol.statistics.get_service_statistics()

    def get_device_registry_statistics(self):
        """Retrieves the IoTHub device registry statistics.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The RegistryStatistics object.
        """
        return self.protocol.statistics.get_device_statistics()

    def get_devices(self, max_number_of_devices=None):
        """Get the identities of multiple devices from the IoTHub identity
           registry. Not recommended. Use the IoTHub query language to retrieve
           device twin and device identity information. See
           https://docs.microsoft.com/en-us/rest/api/iothub/service/queryiothub
           and
           https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-query-language
           for more information.

        :param int max_number_of_devices: This parameter when specified, defines the maximum number
           of device identities that are returned. Any value outside the range of
           1-1000 is considered to be 1000

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: List of device info.
        """
        return self.protocol.devices.get_devices(max_number_of_devices)

    def bulk_create_or_update_devices(self, devices):
        """Create, update, or delete the identities of multiple devices from the
           IoTHub identity registry.

           Create, update, or delete the identities of multiple devices from the
           IoTHub identity registry. A device identity can be specified only once
           in the list. Different operations (create, update, delete) on different
           devices are allowed. A maximum of 100 devices can be specified per
           invocation. For large scale operations, consider using the import
           feature using blob
           storage(https://docs.microsoft.com/azure/iot-hub/iot-hub-devguide-identity-registry#import-and-export-device-identities).

        :param list[ExportImportDevice] devices: The list of device objects to operate on.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The BulkRegistryOperationResult object.
        """
        return self.protocol.bulk_registry.update_registry(devices)

    def query_iot_hub(self, query_specification, continuation_token=None, max_item_count=None):
        """Query an IoTHub to retrieve information regarding device twins using a
           SQL-like language.
           See https://docs.microsoft.com/azure/iot-hub/iot-hub-devguide-query-language
           for more information. Pagination of results is supported. This returns
           information about device twins only.

        :param QuerySpecification query: The query specification.
        :param str continuation_token: Continuation token for paging
        :param str max_item_count: Maximum number of requested device twins

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The QueryResult object.
        """
        raw_response = self.protocol.query.get_twins(
            query_specification, continuation_token, max_item_count, None, True
        )

        queryResult = QueryResult()
        if raw_response.headers:
            queryResult.type = raw_response.headers["x-ms-item-type"]
            queryResult.continuation_token = raw_response.headers["x-ms-continuation"]
        queryResult.items = raw_response.output

        return queryResult

    def get_twin(self, device_id):
        """Gets a device twin.

        :param str device_id: The name (Id) of the device.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        return self.protocol.devices.get_twin(device_id)

    def replace_twin(self, device_id, device_twin, etag=None):
        """Replaces tags and desired properties of a device twin.

        :param str device_id: The name (Id) of the device.
        :param Twin device_twin: The twin info of the device.
        :param str etag: The etag (if_match) value to use for the replace operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        if etag is None:
            etag = "*"

        return self.protocol.devices.replace_twin(device_id, device_twin, _ensure_quoted(etag))

    def update_twin(self, device_id, device_twin, etag=None):
        """Updates tags and desired properties of a device twin.

        :param str device_id: The name (Id) of the device.
        :param Twin device_twin: The twin info of the device.
        :param str etag: The etag (if_match) value to use for the update operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        if etag is None:
            etag = "*"

        return self.protocol.devices.update_twin(device_id, device_twin, _ensure_quoted(etag))

    def get_module_twin(self, device_id, module_id):
        """Gets a module twin.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        return self.protocol.modules.get_twin(device_id, module_id)

    def replace_module_twin(self, device_id, module_id, module_twin, etag=None):
        """Replaces tags and desired properties of a module twin.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param Twin module_twin: The twin info of the module.
        :param str etag: The etag (if_match) value to use for the replace operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        if etag is None:
            etag = "*"

        return self.protocol.modules.replace_twin(
            device_id, module_id, module_twin, _ensure_quoted(etag)
        )

    def update_module_twin(self, device_id, module_id, module_twin, etag=None):
        """Updates tags and desired properties of a module twin.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param Twin module_twin: The twin info of the module.
        :param str etag: The etag (if_match) value to use for the update operation.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The Twin object.
        """
        if etag is None:
            etag = "*"

        return self.protocol.modules.update_twin(
            device_id, module_id, module_twin, _ensure_quoted(etag)
        )

    def invoke_device_method(self, device_id, direct_method_request):
        """Invoke a direct method on a device.

        :param str device_id: The name (Id) of the device.
        :param CloudToDeviceMethod direct_method_request: The method request.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The CloudToDeviceMethodResult object.
        """
        if direct_method_request.payload is None:
            direct_method_request.payload = ""

        return self.protocol.devices.invoke_method(device_id, direct_method_request)

    def invoke_device_module_method(self, device_id, module_id, direct_method_request):
        """Invoke a direct method on a device.

        :param str device_id: The name (Id) of the device.
        :param str module_id: The name (Id) of the module.
        :param CloudToDeviceMethod direct_method_request: The method request.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The CloudToDeviceMethodResult object.
        """
        if direct_method_request.payload is None:
            direct_method_request.payload = ""

        return self.protocol.modules.invoke_method(device_id, module_id, direct_method_request)

    def send_c2d_message(self, device_id, message, properties={}):
        """Send a C2D mesage to a IoTHub Device.

        :param str device_id: The name (Id) of the device.
        :param str message: The message that is to be delievered to the device.
        :param dict properties: The properties to be send with the message.  Can contain
            application properties and system properties

        :raises: Exception if the Send command is not able to send the message
        """

        if self.amqp_svc_client:
            self.amqp_svc_client.send_message_to_device(device_id, message, properties)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from .auth import ConnectionStringAuthentication, AzureIdentityCredentialAdapter
from .protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs as protocol_client


class IoTHubJobManager(object):
    """A class to provide convenience APIs for IoTHub Job Manager operations,
    based on top of the auto generated IotHub REST APIs
    """

    def __init__(self, connection_string=None, host=None, auth=None):
        """Initializer for a Job Manager Service client.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub if we are using connection_str authentication. Default value: None
        :param str host: The Azure service url if we are using token credential authentication.
            Default value: None
        :param str auth: The Azure authentication object if we are using token credential authentication.
            Default value: None

        :returns: Instance of the IoTHubJobManager object.
        :rtype: :class:`azure.iot.hub.IoTHubJobManager`
        """
        if connection_string is not None:
            self.auth = ConnectionStringAuthentication(connection_string)
            self.protocol = protocol_client(self.auth, "https://" + self.auth["HostName"])
        else:
            self.auth = auth
            self.protocol = protocol_client(self.auth, "https://" + host)

    @classmethod
    def from_connection_string(cls, connection_string):
        """Classmethod initializer for a IoTHubJobManager Service client.
        Creates IoTHubJobManager class from connection string.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str connection_string: The IoTHub connection string used to authenticate connection
            with IoTHub.

        :rtype: :class:`azure.iot.hub.IoTHubJobManager`
        """
        return cls(connection_string=connection_string)

    @classmethod
    def from_token_credential(cls, url, token_credential):
        """Classmethod initializer for a IoTHubJobManager Service client.
        Creates IoTHubJobManager class from host name url and Azure token credential.

        After a successful creation the class has been authenticated with IoTHub and
        it is ready to call the member APIs to communicate with IoTHub.

        :param str url: The Azure service url (host name).
        :param str token_credential: The Azure token credential object.

        :rtype: :class:`azure.iot.hub.IoTHubJobManager`
        """
        host = url
        auth = AzureIdentityCredentialAdapter(token_credential)
        return cls(host=host, auth=auth)

    def create_import_export_job(self, job_properties):
        """Creates a new import/export job on an IoT hub.

        :param job_properties: Specifies the job specification.
        :type job_properties: :class:`azure.iot.hub.models.JobProperties`

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: JobProperties object containing the created job.
        :rtype: :class:`azure.iot.hub.models.JobProperties`
        """
        return self.protocol.jobs.create_import_export_job(job_properties)

    def get_import_export_jobs(self):
        """Retrieves the status of all import/export jobs on an IoTHub.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The list[job_properties] object.
        :rtype: :class:`list[azure.iot.hub.models.JobProperties]`
        """
        return self.protocol.jobs.get_import_export_jobs()

    def get_import_export_job(self, job_id):
        """Retrieves the status of an import/export job on an IoTHub.

        :param str job_id: The ID of the job.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The JobProperties object containing the requested job.
        :rtype: :class:`azure.iot.hub.models.JobProperties`
        """
        return self.protocol.jobs.get_import_export_job(job_id)

    def cancel_import_export_job(self, job_id):
        """Cancels an import/export job on an IoT hub.

        :param str job_id: The ID of the job.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: Object.
        """
        return self.protocol.jobs.cancel_import_export_job(job_id)

    def create_scheduled_job(self, job_id, job_request):
        """Creates a new job to schedule update twins or device direct methods on an IoT hub.

        :param str job_id: The ID of the job.
        :param job_request: Specifies the job.
        :type job_request: :class:`azure.iot.hub.models.JobRequest`

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: JobResponse object containing the created job.
        :rtype: :class:`azure.iot.hub.models.JobResponse`
        """
        return self.protocol.jobs.create_scheduled_job(job_id, job_request)

    def get_scheduled_job(self, job_id):
        """Retrieves the details of a scheduled job on an IoTHub.

        :param str job_id: The ID of the job.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: The JobResponse object containing the requested details.
        :rtype: :class:`azure.iot.hub.models.JobResponse`
        """
        return self.protocol.jobs.get_scheduled_job(job_id)

    def cancel_scheduled_job(self, job_id):
        """Cancels a scheduled job on an IoT hub.

        :param str job_id: The ID of the job.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: JobResponse object containing the cancelled job.
        :rtype: :class:`azure.iot.hub.models.JobResponse`
        """
        return self.protocol.jobs.cancel_scheduled_job(job_id)

    def query_scheduled_jobs(self, job_type, job_status):
        """Query an IoT hub to retrieve information regarding jobs using the IoT Hub query language.

        :param str job_type: The type of the jobs.
        :param str job_status: The status of the jobs.

        :raises: `HttpOperationError<msrest.exceptions.HttpOperationError>`
            if the HTTP response status is not in [200].

        :returns: QueryResult object containing the jobs.
        :rtype: :class:`azure.iot.hub.models.QueryResult`
        """
        return self.protocol.jobs.query_scheduled_jobs(job_type, job_status)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class DeviceRegistryOperationError(Model):
    """The device registry operation error details.

    :param device_id: The unique identifier of the device.
    :type device_id: str
    :param error_code: The error code. Possible values include:
     'InvalidErrorCode', 'GenericBadRequest', 'InvalidProtocolVersion',
     'DeviceInvalidResultCount', 'InvalidOperation', 'ArgumentInvalid',
     'ArgumentNull', 'IotHubFormatError',
     'DeviceStorageEntitySerializationError', 'BlobContainerValidationError',
     'ImportWarningExistsError', 'InvalidSchemaVersion',
     'DeviceDefinedMultipleTimes', 'DeserializationError',
     'BulkRegistryOperationFailure', 'DefaultStorageEndpointNotConfigured',
     'InvalidFileUploadCorrelationId', 'ExpiredFileUploadCorrelationId',
     'InvalidStorageEndpoint', 'InvalidMessagingEndpoint',
     'InvalidFileUploadCompletionStatus', 'InvalidStorageEndpointOrBlob',
     'RequestCanceled', 'InvalidStorageEndpointProperty', 'EtagDoesNotMatch',
     'RequestTimedOut', 'UnsupportedOperationOnReplica', 'NullMessage',
     'ConnectionForcefullyClosedOnNewConnection', 'InvalidDeviceScope',
     'ConnectionForcefullyClosedOnFaultInjection',
     'ConnectionRejectedOnFaultInjection', 'InvalidEndpointAuthenticationType',
     'ManagedIdentityNotEnabled', 'InvalidRouteTestInput',
     'InvalidSourceOnRoute', 'RoutingNotEnabled',
     'InvalidContentEncodingOrType', 'InvalidEndorsementKey',
     'InvalidRegistrationId', 'InvalidStorageRootKey',
     'InvalidEnrollmentGroupId', 'TooManyEnrollments',
     'RegistrationIdDefinedMultipleTimes', 'CustomAllocationFailed',
     'CustomAllocationIotHubNotSpecified',
     'CustomAllocationUnauthorizedAccess', 'CannotRegisterModuleToModule',
     'TenantHubRoutingNotEnabled', 'InvalidConfigurationTargetCondition',
     'InvalidConfigurationContent',
     'CannotModifyImmutableConfigurationContent',
     'InvalidConfigurationCustomMetricsQuery', 'InvalidPnPInterfaceDefinition',
     'InvalidPnPDesiredProperties', 'InvalidPnPReportedProperties',
     'InvalidPnPWritableReportedProperties', 'InvalidDigitalTwinJsonPatch',
     'InvalidDigitalTwinPayload', 'InvalidDigitalTwinPatch',
     'InvalidDigitalTwinPatchPath', 'GenericUnauthorized', 'IotHubNotFound',
     'IotHubUnauthorizedAccess', 'IotHubUnauthorized', 'ElasticPoolNotFound',
     'SystemModuleModifyUnauthorizedAccess', 'GenericForbidden',
     'IotHubSuspended', 'IotHubQuotaExceeded', 'JobQuotaExceeded',
     'DeviceMaximumQueueDepthExceeded', 'IotHubMaxCbsTokenExceeded',
     'DeviceMaximumActiveFileUploadLimitExceeded',
     'DeviceMaximumQueueSizeExceeded', 'RoutingEndpointResponseForbidden',
     'InvalidMessageExpiryTime', 'OperationNotAvailableInCurrentTier',
     'KeyEncryptionKeyRevoked', 'DeviceModelMaxPropertiesExceeded',
     'DeviceModelMaxIndexablePropertiesExceeded', 'IotDpsSuspended',
     'IotDpsSuspending', 'GenericNotFound', 'DeviceNotFound', 'JobNotFound',
     'QuotaMetricNotFound', 'SystemPropertyNotFound', 'AmqpAddressNotFound',
     'RoutingEndpointResponseNotFound', 'CertificateNotFound',
     'ElasticPoolTenantHubNotFound', 'ModuleNotFound',
     'AzureTableStoreNotFound', 'IotHubFailingOver', 'FeatureNotSupported',
     'DigitalTwinInterfaceNotFound', 'QueryStoreClusterNotFound',
     'DeviceNotOnline', 'DeviceConnectionClosedRemotely', 'EnrollmentNotFound',
     'DeviceRegistrationNotFound', 'AsyncOperationNotFound',
     'EnrollmentGroupNotFound', 'DeviceRecordNotFound', 'GroupRecordNotFound',
     'DeviceGroupNotFound', 'ProvisioningSettingsNotFound',
     'ProvisioningRecordNotFound', 'LinkedHubNotFound',
     'CertificateAuthorityNotFound', 'ConfigurationNotFound', 'GroupNotFound',
     'DigitalTwinModelNotFound', 'InterfaceNameModelNotFound',
     'GenericMethodNotAllowed', 'OperationNotAllowedInCurrentState',
     'ImportDevicesNotSupported', 'BulkAddDevicesNotSupported',
     'GenericConflict', 'DeviceAlreadyExists', 'LinkCreationConflict',
     'CallbackSubscriptionConflict', 'ModelAlreadyExists', 'DeviceLocked',
     'DeviceJobAlreadyExists', 'JobAlreadyExists', 'EnrollmentConflict',
     'EnrollmentGroupConflict', 'RegistrationStatusConflict',
     'DeviceRecordConflict', 'GroupRecordConflict', 'DeviceGroupConflict',
     'ProvisioningSettingsConflict', 'ProvisioningRecordConflict',
     'LinkedHubConflict', 'CertificateAuthorityConflict',
     'ModuleAlreadyExistsOnDevice', 'ConfigurationAlreadyExists',
     'ApplyConfigurationAlreadyInProgressOnDevice',
     'DigitalTwinModelAlreadyExists',
     'DigitalTwinModelExistsWithOtherModelType',
     'InterfaceNameModelAlreadyExists', 'GenericPreconditionFailed',
     'PreconditionFailed', 'DeviceMessageLockLost', 'JobRunPreconditionFailed',
     'InflightMessagesInLink', 'GenericRequestEntityTooLarge',
     'MessageTooLarge', 'TooManyDevices', 'TooManyModulesOnDevice',
     'ConfigurationCountLimitExceeded', 'DigitalTwinModelCountLimitExceeded',
     'InterfaceNameCompressionModelCountLimitExceeded',
     'GenericUnsupportedMediaType', 'IncompatibleDataType',
     'GenericTooManyRequests', 'ThrottlingException',
     'ThrottleBacklogLimitExceeded', 'ThrottlingBacklogTimeout',
     'ThrottlingMaxActiveJobCountExceeded', 'DeviceThrottlingLimitExceeded',
     'ClientClosedRequest', 'GenericServerError', 'ServerError',
     'JobCancelled', 'StatisticsRetrievalError', 'ConnectionForcefullyClosed',
     'InvalidBlobState', 'BackupTimedOut', 'AzureStorageTimeout',
     'GenericTimeout', 'InvalidThrottleParameter', 'EventHubLinkAlreadyClosed',
     'ReliableBlobStoreError', 'RetryAttemptsExhausted',
     'AzureTableStoreError', 'CheckpointStoreNotFound',
     'DocumentDbInvalidReturnValue', 'ReliableDocDbStoreStoreError',
     'ReliableBlobStoreTimeoutError', 'ConfigReadFailed',
     'InvalidContainerReceiveLink', 'InvalidPartitionEpoch', 'RestoreTimedOut',
     'StreamReservationFailure', 'SerializationError',
     'UnexpectedPropertyValue', 'OrchestrationOperationFailed',
     'ModelRepoEndpointError', 'ResolutionError', 'UnableToFetchCredentials',
     'UnableToFetchTenantInfo', 'UnableToShareIdentity',
     'UnableToExpandDiscoveryInfo', 'UnableToExpandComponentInfo',
     'UnableToCompressComponentInfo', 'UnableToCompressDiscoveryInfo',
     'OrphanDiscoveryDocument', 'GenericBadGateway',
     'InvalidResponseWhileProxying', 'GenericServiceUnavailable',
     'ServiceUnavailable', 'PartitionNotFound', 'IotHubActivationFailed',
     'ServerBusy', 'IotHubRestoring', 'ReceiveLinkOpensThrottled',
     'ConnectionUnavailable', 'DeviceUnavailable', 'ConfigurationNotAvailable',
     'GroupNotAvailable', 'HostingServiceNotAvailable',
     'GenericGatewayTimeout', 'GatewayTimeout'
    :type error_code: str or ~protocol.models.enum
    :param error_status: The details of the error.
    :type error_status: str
    :param module_id: The unique identifier of the module, if applicable.
    :type module_id: str
    :param operation: The type of the operation that failed.
    :type operation: str
    """

    _attribute_map = {
        "device_id": {"key": "deviceId", "type": "str"},
        "error_code": {"key": "errorCode", "type": "str"},
        "error_status": {"key": "errorStatus", "type": "str"},
        "module_id": {"key": "moduleId", "type": "str"},
        "operation": {"key": "operation", "type": "str"},
    }

    def __init__(self, **kwargs):
        super(DeviceRegistryOperationError, self).__init__(**kwargs)
        self.device_id = kwargs.get("device_id", None)
        self.error_code = kwargs.get("error_code", None)
        self.error_status = kwargs.get("error_status", None)
        self.module_id = kwargs.get("module_id", None)
        self.operation = kwargs.get("operation", None)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.pipeline import ClientRawResponse
from msrest.exceptions import HttpOperationError

from .. import models


class ConfigurationOperations(object):
    """ConfigurationOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    :ivar api_version: Version of the Api. Constant value: "2021-04-12".
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config
        self.api_version = "2021-04-12"

    def get(self, id, custom_headers=None, raw=False, **operation_config):
        """Gets a configuration on the IoT Hub for automatic device/module
        management.

        :param id: The unique identifier of the configuration.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: Configuration or ClientRawResponse if raw=true
        :rtype: ~protocol.models.Configuration or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("Configuration", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get.metadata = {"url": "/configurations/{id}"}

    def create_or_update(
        self, id, configuration, if_match=None, custom_headers=None, raw=False, **operation_config
    ):
        """Creates or updates a configuration on the IoT Hub for automatic
        device/module management. Configuration identifier and Content cannot
        be updated.

        :param id: The unique identifier of the configuration.
        :type id: str
        :param configuration: The configuration to be created or updated.
        :type configuration: ~protocol.models.Configuration
        :param if_match: The string representing a weak ETag for the
         configuration, as per RFC7232. This should not be set when creating a
         configuration, but may be set when updating a configuration.
        :type if_match: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: Configuration or ClientRawResponse if raw=true
        :rtype: ~protocol.models.Configuration or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.create_or_update.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)
        if if_match is not None:
            header_parameters["If-Match"] = self._serialize.header("if_match", if_match, "str")

        # Construct body
        body_content = self._serialize.body(configuration, "Configuration")

        # Construct and send request
        request = self._client.put(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200, 201]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("Configuration", response)
        if response.status_code == 201:
            deserialized = self._deserialize("Configuration", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    create_or_update.metadata = {"url": "/configurations/{id}"}

    def delete(self, id, if_match=None, custom_headers=None, raw=False, **operation_config):
        """Deletes a configuration on the IoT Hub for automatic device/module
        management.

        :param id: The unique identifier of the configuration.
        :type id: str
        :param if_match: The string representing a weak ETag for the
         configuration, as per RFC7232. The delete operation is performed only
         if this ETag matches the value maintained by the server, indicating
         that the configuration has not been modified since it was last
         retrieved. To force an unconditional delete, set If-Match to the
         wildcard character (*).
        :type if_match: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: None or ClientRawResponse if raw=true
        :rtype: None or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.delete.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        if custom_headers:
            header_parameters.update(custom_headers)
        if if_match is not None:
            header_parameters["If-Match"] = self._serialize.header("if_match", if_match, "str")

        # Construct and send request
        request = self._client.delete(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [204]:
            raise HttpOperationError(self._deserialize, response)

        if raw:
            client_raw_response = ClientRawResponse(None, response)
            return client_raw_response

    delete.metadata = {"url": "/configurations/{id}"}

    def get_configurations(self, top=None, custom_headers=None, raw=False, **operation_config):
        """Gets configurations on the IoT Hub for automatic device/module
        management. Pagination is not supported.

        :param top: The number of configurations to retrieve. Value will be
         overridden if greater than the maximum deployment count for the IoT
         Hub.
        :type top: int
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: list or ClientRawResponse if raw=true
        :rtype: list[~protocol.models.Configuration] or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_configurations.metadata["url"]

        # Construct parameters
        query_parameters = {}
        if top is not None:
            query_parameters["top"] = self._serialize.query("top", top, "int")
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("[Configuration]", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_configurations.metadata = {"url": "/configurations"}

    def test_queries(self, input, custom_headers=None, raw=False, **operation_config):
        """Validates target condition and custom metric queries for a
        configuration on the IoT Hub.

        :param input: The configuration for target condition and custom metric
         queries.
        :type input: ~protocol.models.ConfigurationQueriesTestInput
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: ConfigurationQueriesTestResponse or ClientRawResponse if
         raw=true
        :rtype: ~protocol.models.ConfigurationQueriesTestResponse or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.test_queries.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(input, "ConfigurationQueriesTestInput")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("ConfigurationQueriesTestResponse", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    test_queries.metadata = {"url": "/configurations/testQueries"}

    def apply_on_edge_device(self, id, content, custom_headers=None, raw=False, **operation_config):
        """Applies the configuration content to an edge device.

        :param id: The unique identifier of the edge device.
        :type id: str
        :param content: The configuration content.
        :type content: ~protocol.models.ConfigurationContent
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: None or ClientRawResponse if raw=true
        :rtype: None or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.apply_on_edge_device.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(content, "ConfigurationContent")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [204]:
            raise HttpOperationError(self._deserialize, response)

        if raw:
            client_raw_response = ClientRawResponse(None, response)
            return client_raw_response

    apply_on_edge_device.metadata = {"url": "/devices/{id}/applyConfigurationContent"}



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ExportImportDevice(Model):
    """ExportImportDevice.

    :param id: The unique identifier of the device.
    :type id: str
    :param module_id: The unique identifier of the module, if applicable.
    :type module_id: str
    :param e_tag: The string representing a weak ETag for the device RFC7232.
     The value is only used if import mode is updateIfMatchETag, in that case
     the import operation is performed only if this ETag matches the value
     maintained by the server.
    :type e_tag: str
    :param import_mode: The type of registry operation and ETag preferences.
     Possible values include: 'create', 'update', 'updateIfMatchETag',
     'delete', 'deleteIfMatchETag', 'updateTwin', 'updateTwinIfMatchETag'
    :type import_mode: str or ~protocol.models.enum
    :param status: The status of the module. If disabled, the module cannot
     connect to the service. Possible values include: 'enabled', 'disabled'
    :type status: str or ~protocol.models.enum
    :param status_reason: The 128 character-long string that stores the reason
     for the device identity status. All UTF-8 characters are allowed.
    :type status_reason: str
    :param authentication: The authentication mechanism used by the module.
     This parameter is optional and defaults to SAS if not provided. In that
     case, primary/secondary access keys are auto-generated.
    :type authentication: ~protocol.models.AuthenticationMechanism
    :param twin_etag: The string representing a weak ETag for the device twin
     RFC7232. The value is only used if import mode is updateIfMatchETag, in
     that case the import operation is performed only if this ETag matches the
     value maintained by the server.
    :type twin_etag: str
    :param tags: The JSON document read and written by the solution back end.
     The tags are not visible to device apps.
    :type tags: dict[str, object]
    :param properties: The desired and reported properties for the device.
    :type properties: ~protocol.models.PropertyContainer
    :param capabilities: The status of capabilities enabled on the device.
    :type capabilities: ~protocol.models.DeviceCapabilities
    :param device_scope: The scope of the device.
    :type device_scope: str
    :param parent_scopes: The scopes of the upper level edge devices if
     applicable. Only available for edge devices.
    :type parent_scopes: list[str]
    """

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "module_id": {"key": "moduleId", "type": "str"},
        "e_tag": {"key": "eTag", "type": "str"},
        "import_mode": {"key": "importMode", "type": "str"},
        "status": {"key": "status", "type": "str"},
        "status_reason": {"key": "statusReason", "type": "str"},
        "authentication": {"key": "authentication", "type": "AuthenticationMechanism"},
        "twin_etag": {"key": "twinETag", "type": "str"},
        "tags": {"key": "tags", "type": "{object}"},
        "properties": {"key": "properties", "type": "PropertyContainer"},
        "capabilities": {"key": "capabilities", "type": "DeviceCapabilities"},
        "device_scope": {"key": "deviceScope", "type": "str"},
        "parent_scopes": {"key": "parentScopes", "type": "[str]"},
    }

    def __init__(
        self,
        *,
        id: str = None,
        module_id: str = None,
        e_tag: str = None,
        import_mode=None,
        status=None,
        status_reason: str = None,
        authentication=None,
        twin_etag: str = None,
        tags=None,
        properties=None,
        capabilities=None,
        device_scope: str = None,
        parent_scopes=None,
        **kwargs
    ) -> None:
        super(ExportImportDevice, self).__init__(**kwargs)
        self.id = id
        self.module_id = module_id
        self.e_tag = e_tag
        self.import_mode = import_mode
        self.status = status
        self.status_reason = status_reason
        self.authentication = authentication
        self.twin_etag = twin_etag
        self.tags = tags
        self.properties = properties
        self.capabilities = capabilities
        self.device_scope = device_scope
        self.parent_scopes = parent_scopes



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.pipeline import ClientRawResponse
from msrest.exceptions import HttpOperationError

from .. import models


class DigitalTwinOperations(object):
    """DigitalTwinOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    :ivar api_version: Version of the Api. Constant value: "2021-04-12".
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config
        self.api_version = "2021-04-12"

    def get_digital_twin(self, id, custom_headers=None, raw=False, **operation_config):
        """Gets a digital twin.

        :param id: Digital Twin ID.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: object or ClientRawResponse if raw=true
        :rtype: object or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_digital_twin.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None
        header_dict = {}

        if response.status_code == 200:
            deserialized = self._deserialize("object", response)
            header_dict = {"ETag": "str"}

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            client_raw_response.add_headers(header_dict)
            return client_raw_response

        return deserialized

    get_digital_twin.metadata = {"url": "/digitaltwins/{id}"}

    def update_digital_twin(
        self,
        id,
        digital_twin_patch,
        if_match=None,
        custom_headers=None,
        raw=False,
        **operation_config
    ):
        """Updates a digital twin.

        :param id: Digital Twin ID.
        :type id: str
        :param digital_twin_patch: json-patch contents to update.
        :type digital_twin_patch: list[object]
        :param if_match:
        :type if_match: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: None or ClientRawResponse if raw=true
        :rtype: None or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.update_digital_twin.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)
        if if_match is not None:
            header_parameters["If-Match"] = self._serialize.header("if_match", if_match, "str")

        # Construct body
        body_content = self._serialize.body(digital_twin_patch, "[object]")

        # Construct and send request
        request = self._client.patch(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [202]:
            raise HttpOperationError(self._deserialize, response)

        if raw:
            client_raw_response = ClientRawResponse(None, response)
            client_raw_response.add_headers({"ETag": "str", "Location": "str"})
            return client_raw_response

    update_digital_twin.metadata = {"url": "/digitaltwins/{id}"}

    def invoke_root_level_command(
        self,
        id,
        command_name,
        payload,
        connect_timeout_in_seconds=None,
        response_timeout_in_seconds=None,
        custom_headers=None,
        raw=False,
        **operation_config
    ):
        """Invoke a digital twin root level command.

        Invoke a digital twin root level command.

        :param id:
        :type id: str
        :param command_name:
        :type command_name: str
        :param payload:
        :type payload: object
        :param connect_timeout_in_seconds: Maximum interval of time, in
         seconds, that the digital twin command will wait for the answer.
        :type connect_timeout_in_seconds: int
        :param response_timeout_in_seconds: Maximum interval of time, in
         seconds, that the digital twin command will wait for the answer.
        :type response_timeout_in_seconds: int
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: object or ClientRawResponse if raw=true
        :rtype: object or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.invoke_root_level_command.metadata["url"]
        path_format_arguments = {
            "id": self._serialize.url("id", id, "str"),
            "commandName": self._serialize.url("command_name", command_name, "str"),
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )
        if connect_timeout_in_seconds is not None:
            query_parameters["connectTimeoutInSeconds"] = self._serialize.query(
                "connect_timeout_in_seconds", connect_timeout_in_seconds, "int"
            )
        if response_timeout_in_seconds is not None:
            query_parameters["responseTimeoutInSeconds"] = self._serialize.query(
                "response_timeout_in_seconds", response_timeout_in_seconds, "int"
            )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(payload, "object")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None
        header_dict = {}

        if response.status_code == 200:
            deserialized = self._deserialize("object", response)
            header_dict = {"x-ms-command-statuscode": "int", "x-ms-request-id": "str"}

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            client_raw_response.add_headers(header_dict)
            return client_raw_response

        return deserialized

    invoke_root_level_command.metadata = {"url": "/digitaltwins/{id}/commands/{commandName}"}

    def invoke_component_command(
        self,
        id,
        component_path,
        command_name,
        payload,
        connect_timeout_in_seconds=None,
        response_timeout_in_seconds=None,
        custom_headers=None,
        raw=False,
        **operation_config
    ):
        """Invoke a digital twin command.

        Invoke a digital twin command.

        :param id:
        :type id: str
        :param component_path:
        :type component_path: str
        :param command_name:
        :type command_name: str
        :param payload:
        :type payload: object
        :param connect_timeout_in_seconds: Maximum interval of time, in
         seconds, that the digital twin command will wait for the answer.
        :type connect_timeout_in_seconds: int
        :param response_timeout_in_seconds: Maximum interval of time, in
         seconds, that the digital twin command will wait for the answer.
        :type response_timeout_in_seconds: int
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: object or ClientRawResponse if raw=true
        :rtype: object or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.invoke_component_command.metadata["url"]
        path_format_arguments = {
            "id": self._serialize.url("id", id, "str"),
            "componentPath": self._serialize.url("component_path", component_path, "str"),
            "commandName": self._serialize.url("command_name", command_name, "str"),
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )
        if connect_timeout_in_seconds is not None:
            query_parameters["connectTimeoutInSeconds"] = self._serialize.query(
                "connect_timeout_in_seconds", connect_timeout_in_seconds, "int"
            )
        if response_timeout_in_seconds is not None:
            query_parameters["responseTimeoutInSeconds"] = self._serialize.query(
                "response_timeout_in_seconds", response_timeout_in_seconds, "int"
            )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(payload, "object")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None
        header_dict = {}

        if response.status_code == 200:
            deserialized = self._deserialize("object", response)
            header_dict = {"x-ms-command-statuscode": "int", "x-ms-request-id": "str"}

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            client_raw_response.add_headers(header_dict)
            return client_raw_response

        return deserialized

    invoke_component_command.metadata = {
        "url": "/digitaltwins/{id}/components/{componentPath}/commands/{commandName}"
    }



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.service_client import SDKClient
from msrest import Configuration, Serializer, Deserializer
from .version import VERSION
from msrest.exceptions import HttpOperationError
from .operations.configuration_operations import ConfigurationOperations
from .operations.statistics_operations import StatisticsOperations
from .operations.devices_operations import DevicesOperations
from .operations.bulk_registry_operations import BulkRegistryOperations
from .operations.query_operations import QueryOperations
from .operations.jobs_operations import JobsOperations
from .operations.cloud_to_device_messages_operations import CloudToDeviceMessagesOperations
from .operations.modules_operations import ModulesOperations
from .operations.digital_twin_operations import DigitalTwinOperations
from . import models


class IotHubGatewayServiceAPIsConfiguration(Configuration):
    """Configuration for IotHubGatewayServiceAPIs
    Note that all parameters used to create this instance are saved as instance
    attributes.

    :param credentials: Subscription credentials which uniquely identify
     client subscription.
    :type credentials: None
    :param str base_url: Service URL
    """

    def __init__(self, credentials, base_url=None):

        if credentials is None:
            raise ValueError("Parameter 'credentials' must not be None.")
        if not base_url:
            base_url = "https://fully-qualified-iothubname.azure-devices.net"

        super(IotHubGatewayServiceAPIsConfiguration, self).__init__(base_url)

        self.add_user_agent("iothubgatewayserviceapis/{}".format(VERSION))

        self.credentials = credentials


class IotHubGatewayServiceAPIs(SDKClient):
    """IotHubGatewayServiceAPIs

    :ivar config: Configuration for client.
    :vartype config: IotHubGatewayServiceAPIsConfiguration

    :ivar configuration: Configuration operations
    :vartype configuration: protocol.operations.ConfigurationOperations
    :ivar statistics: Statistics operations
    :vartype statistics: protocol.operations.StatisticsOperations
    :ivar devices: Devices operations
    :vartype devices: protocol.operations.DevicesOperations
    :ivar bulk_registry: BulkRegistry operations
    :vartype bulk_registry: protocol.operations.BulkRegistryOperations
    :ivar query: Query operations
    :vartype query: protocol.operations.QueryOperations
    :ivar jobs: Jobs operations
    :vartype jobs: protocol.operations.JobsOperations
    :ivar cloud_to_device_messages: CloudToDeviceMessages operations
    :vartype cloud_to_device_messages: protocol.operations.CloudToDeviceMessagesOperations
    :ivar modules: Modules operations
    :vartype modules: protocol.operations.ModulesOperations
    :ivar digital_twin: DigitalTwin operations
    :vartype digital_twin: protocol.operations.DigitalTwinOperations

    :param credentials: Subscription credentials which uniquely identify
     client subscription.
    :type credentials: None
    :param str base_url: Service URL
    """

    def __init__(self, credentials, base_url=None):

        self.config = IotHubGatewayServiceAPIsConfiguration(credentials, base_url)
        super(IotHubGatewayServiceAPIs, self).__init__(self.config.credentials, self.config)

        client_models = {k: v for k, v in models.__dict__.items() if isinstance(v, type)}
        self.api_version = "2021-04-12"
        self._serialize = Serializer(client_models)
        self._deserialize = Deserializer(client_models)

        self.configuration = ConfigurationOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.statistics = StatisticsOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.devices = DevicesOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.bulk_registry = BulkRegistryOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.query = QueryOperations(self._client, self.config, self._serialize, self._deserialize)
        self.jobs = JobsOperations(self._client, self.config, self._serialize, self._deserialize)
        self.cloud_to_device_messages = CloudToDeviceMessagesOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.modules = ModulesOperations(
            self._client, self.config, self._serialize, self._deserialize
        )
        self.digital_twin = DigitalTwinOperations(
            self._client, self.config, self._serialize, self._deserialize
        )



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class JobResponse(Model):
    """JobResponse.

    :param job_id: System generated.  Ignored at creation. The unique
     identifier of the job.
    :type job_id: str
    :param query_condition: The device query condition.
    :type query_condition: str
    :param created_time: System generated.  Ignored at creation. The creation
     date and time of the job.
    :type created_time: datetime
    :param start_time: The start date and time of the scheduled job in UTC.
    :type start_time: datetime
    :param end_time: System generated.  Ignored at creation. The end date and
     time of the job in UTC.
    :type end_time: datetime
    :param max_execution_time_in_seconds: The maximum execution time in
     secounds.
    :type max_execution_time_in_seconds: long
    :param type: The job type. Possible values include: 'unknown', 'export',
     'import', 'backup', 'readDeviceProperties', 'writeDeviceProperties',
     'updateDeviceConfiguration', 'rebootDevice', 'factoryResetDevice',
     'firmwareUpdate', 'scheduleDeviceMethod', 'scheduleUpdateTwin',
     'restoreFromBackup', 'failoverDataCopy'
    :type type: str or ~protocol.models.enum
    :param cloud_to_device_method: The method type and parameters. This is
     required if job type is cloudToDeviceMethod.
    :type cloud_to_device_method: ~protocol.models.CloudToDeviceMethod
    :param update_twin:
    :type update_twin: ~protocol.models.Twin
    :param status: System generated.  Ignored at creation. The status of the
     job. Possible values include: 'unknown', 'enqueued', 'running',
     'completed', 'failed', 'cancelled', 'scheduled', 'queued'
    :type status: str or ~protocol.models.enum
    :param failure_reason: The reason for the failure, if a failure occurred.
    :type failure_reason: str
    :param status_message: The status message of the job.
    :type status_message: str
    :param device_job_statistics: The details regarding job execution status.
    :type device_job_statistics: ~protocol.models.DeviceJobStatistics
    """

    _attribute_map = {
        "job_id": {"key": "jobId", "type": "str"},
        "query_condition": {"key": "queryCondition", "type": "str"},
        "created_time": {"key": "createdTime", "type": "iso-8601"},
        "start_time": {"key": "startTime", "type": "iso-8601"},
        "end_time": {"key": "endTime", "type": "iso-8601"},
        "max_execution_time_in_seconds": {"key": "maxExecutionTimeInSeconds", "type": "long"},
        "type": {"key": "type", "type": "str"},
        "cloud_to_device_method": {"key": "cloudToDeviceMethod", "type": "CloudToDeviceMethod"},
        "update_twin": {"key": "updateTwin", "type": "Twin"},
        "status": {"key": "status", "type": "str"},
        "failure_reason": {"key": "failureReason", "type": "str"},
        "status_message": {"key": "statusMessage", "type": "str"},
        "device_job_statistics": {"key": "deviceJobStatistics", "type": "DeviceJobStatistics"},
    }

    def __init__(
        self,
        *,
        job_id: str = None,
        query_condition: str = None,
        created_time=None,
        start_time=None,
        end_time=None,
        max_execution_time_in_seconds: int = None,
        type=None,
        cloud_to_device_method=None,
        update_twin=None,
        status=None,
        failure_reason: str = None,
        status_message: str = None,
        device_job_statistics=None,
        **kwargs
    ) -> None:
        super(JobResponse, self).__init__(**kwargs)
        self.job_id = job_id
        self.query_condition = query_condition
        self.created_time = created_time
        self.start_time = start_time
        self.end_time = end_time
        self.max_execution_time_in_seconds = max_execution_time_in_seconds
        self.type = type
        self.cloud_to_device_method = cloud_to_device_method
        self.update_twin = update_twin
        self.status = status
        self.failure_reason = failure_reason
        self.status_message = status_message
        self.device_job_statistics = device_job_statistics



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ManagedIdentity(Model):
    """ManagedIdentity.

    :param user_assigned_identity:
    :type user_assigned_identity: str
    """

    _attribute_map = {"user_assigned_identity": {"key": "userAssignedIdentity", "type": "str"}}

    def __init__(self, *, user_assigned_identity: str = None, **kwargs) -> None:
        super(ManagedIdentity, self).__init__(**kwargs)
        self.user_assigned_identity = user_assigned_identity



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.pipeline import ClientRawResponse
from msrest.exceptions import HttpOperationError

from .. import models


class JobsOperations(object):
    """JobsOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    :ivar api_version: Version of the Api. Constant value: "2021-04-12".
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config
        self.api_version = "2021-04-12"

    def create_import_export_job(
        self, job_properties, custom_headers=None, raw=False, **operation_config
    ):
        """Creates a new import or export job on the IoT Hub. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-identity-registry#import-and-export-device-identities
        for more information.

        :param job_properties: The job specifications.
        :type job_properties: ~protocol.models.JobProperties
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobProperties or ClientRawResponse if raw=true
        :rtype: ~protocol.models.JobProperties or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.create_import_export_job.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(job_properties, "JobProperties")

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("JobProperties", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    create_import_export_job.metadata = {"url": "/jobs/create"}

    def get_import_export_jobs(self, custom_headers=None, raw=False, **operation_config):
        """Gets the status of all import and export jobs in the IoT Hub. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-identity-registry#import-and-export-device-identities
        for more information.

        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: list or ClientRawResponse if raw=true
        :rtype: list[~protocol.models.JobProperties] or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_import_export_jobs.metadata["url"]

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("[JobProperties]", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_import_export_jobs.metadata = {"url": "/jobs"}

    def get_import_export_job(self, id, custom_headers=None, raw=False, **operation_config):
        """Gets the status of an import or export job in the IoT Hub. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-identity-registry#import-and-export-device-identities
        for more information.

        :param id: The unique identifier of the job.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobProperties or ClientRawResponse if raw=true
        :rtype: ~protocol.models.JobProperties or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_import_export_job.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("JobProperties", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_import_export_job.metadata = {"url": "/jobs/{id}"}

    def cancel_import_export_job(self, id, custom_headers=None, raw=False, **operation_config):
        """Cancels an import or export job in the IoT Hub.

        :param id: The unique identifier of the job.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: object or ClientRawResponse if raw=true
        :rtype: object or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.cancel_import_export_job.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.delete(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200, 204]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("object", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    cancel_import_export_job.metadata = {"url": "/jobs/{id}"}

    def get_scheduled_job(self, id, custom_headers=None, raw=False, **operation_config):
        """Gets details of a scheduled job from the IoT Hub. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-jobs
        for more information.

        :param id: The unique identifier of the job.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobResponse or ClientRawResponse if raw=true
        :rtype: ~protocol.models.JobResponse or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.get_scheduled_job.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("JobResponse", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    get_scheduled_job.metadata = {"url": "/jobs/v2/{id}"}

    def create_scheduled_job(
        self, id, job_request, custom_headers=None, raw=False, **operation_config
    ):
        """Creates a new job to schedule twin updates or direct methods on the IoT
        Hub at a scheduled time. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-jobs
        for more information.

        :param id: The unique identifier of the job.
        :type id: str
        :param job_request: The job request info.
        :type job_request: ~protocol.models.JobRequest
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobResponse or ClientRawResponse if raw=true
        :rtype: ~protocol.models.JobResponse or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.create_scheduled_job.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        header_parameters["Content-Type"] = "application/json; charset=utf-8"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct body
        body_content = self._serialize.body(job_request, "JobRequest")

        # Construct and send request
        request = self._client.put(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("JobResponse", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    create_scheduled_job.metadata = {"url": "/jobs/v2/{id}"}

    def cancel_scheduled_job(self, id, custom_headers=None, raw=False, **operation_config):
        """Cancels a scheduled job on the IoT Hub. See
        https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-jobs
        for more information.

        :param id: The unique identifier of the job.
        :type id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobResponse or ClientRawResponse if raw=true
        :rtype: ~protocol.models.JobResponse or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.cancel_scheduled_job.metadata["url"]
        path_format_arguments = {"id": self._serialize.url("id", id, "str")}
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("JobResponse", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    cancel_scheduled_job.metadata = {"url": "/jobs/v2/{id}/cancel"}

    def query_scheduled_jobs(
        self, job_type=None, job_status=None, custom_headers=None, raw=False, **operation_config
    ):
        """Gets the information about jobs using an IoT Hub query. See
        https://docs.microsoft.com/azure/iot-hub/iot-hub-devguide-query-language
        for more information.

        :param job_type: The job type. See
         https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-jobs#querying-for-progress-on-jobs
         for a list of possible job types.
        :type job_type: str
        :param job_status: The job status. See
         https://docs.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-jobs#querying-for-progress-on-jobs
         for a list of possible statuses.
        :type job_status: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: QueryResult or ClientRawResponse if raw=true
        :rtype: ~protocol.models.QueryResult or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`HttpOperationError<msrest.exceptions.HttpOperationError>`
        """
        # Construct URL
        url = self.query_scheduled_jobs.metadata["url"]

        # Construct parameters
        query_parameters = {}
        if job_type is not None:
            query_parameters["jobType"] = self._serialize.query("job_type", job_type, "str")
        if job_status is not None:
            query_parameters["jobStatus"] = self._serialize.query("job_status", job_status, "str")
        query_parameters["api-version"] = self._serialize.query(
            "self.api_version", self.api_version, "str"
        )

        # Construct headers
        header_parameters = {}
        header_parameters["Accept"] = "application/json"
        if custom_headers:
            header_parameters.update(custom_headers)

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise HttpOperationError(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize("QueryResult", response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized

    query_scheduled_jobs.metadata = {"url": "/jobs/v2/query"}



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class Device(Model):
    """Device.

    :param device_id: The unique identifier of the device.
    :type device_id: str
    :param generation_id: The IoT Hub-generated, case-sensitive string up to
     128 characters long. This value is used to distinguish devices with the
     same deviceId, when they have been deleted and re-created.
    :type generation_id: str
    :param etag: The string representing a weak ETag for the device identity,
     as per RFC7232.
    :type etag: str
    :param connection_state: The state of the device. Possible values include:
     'Disconnected', 'Connected'
    :type connection_state: str or ~protocol.models.enum
    :param status: The status of the device. If the status disabled, a device
     cannot connect to the service. Possible values include: 'enabled',
     'disabled'
    :type status: str or ~protocol.models.enum
    :param status_reason: The 128 character-long string that stores the reason
     for the device identity status. All UTF-8 characters are allowed.
    :type status_reason: str
    :param connection_state_updated_time: The date and time the connection
     state was last updated.
    :type connection_state_updated_time: datetime
    :param status_updated_time: The date and time when the status field was
     last updated.
    :type status_updated_time: datetime
    :param last_activity_time: The date and last time the device last
     connected, received, or sent a message.
    :type last_activity_time: datetime
    :param cloud_to_device_message_count: The number of cloud-to-device
     messages currently queued to be sent to the device.
    :type cloud_to_device_message_count: int
    :param authentication: The authentication mechanism used by the device.
    :type authentication: ~protocol.models.AuthenticationMechanism
    :param capabilities: The set of capabilities of the device. For example,
     if this device is an edge device or not.
    :type capabilities: ~protocol.models.DeviceCapabilities
    :param device_scope: The scope of the device. Auto generated and immutable
     for edge devices and modifiable in leaf devices to create child/parent
     relationship.
    :type device_scope: str
    :param parent_scopes: The scopes of the upper level edge devices if
     applicable. Only available for edge devices.
    :type parent_scopes: list[str]
    """

    _attribute_map = {
        "device_id": {"key": "deviceId", "type": "str"},
        "generation_id": {"key": "generationId", "type": "str"},
        "etag": {"key": "etag", "type": "str"},
        "connection_state": {"key": "connectionState", "type": "str"},
        "status": {"key": "status", "type": "str"},
        "status_reason": {"key": "statusReason", "type": "str"},
        "connection_state_updated_time": {"key": "connectionStateUpdatedTime", "type": "iso-8601"},
        "status_updated_time": {"key": "statusUpdatedTime", "type": "iso-8601"},
        "last_activity_time": {"key": "lastActivityTime", "type": "iso-8601"},
        "cloud_to_device_message_count": {"key": "cloudToDeviceMessageCount", "type": "int"},
        "authentication": {"key": "authentication", "type": "AuthenticationMechanism"},
        "capabilities": {"key": "capabilities", "type": "DeviceCapabilities"},
        "device_scope": {"key": "deviceScope", "type": "str"},
        "parent_scopes": {"key": "parentScopes", "type": "[str]"},
    }

    def __init__(self, **kwargs):
        super(Device, self).__init__(**kwargs)
        self.device_id = kwargs.get("device_id", None)
        self.generation_id = kwargs.get("generation_id", None)
        self.etag = kwargs.get("etag", None)
        self.connection_state = kwargs.get("connection_state", None)
        self.status = kwargs.get("status", None)
        self.status_reason = kwargs.get("status_reason", None)
        self.connection_state_updated_time = kwargs.get("connection_state_updated_time", None)
        self.status_updated_time = kwargs.get("status_updated_time", None)
        self.last_activity_time = kwargs.get("last_activity_time", None)
        self.cloud_to_device_message_count = kwargs.get("cloud_to_device_message_count", None)
        self.authentication = kwargs.get("authentication", None)
        self.capabilities = kwargs.get("capabilities", None)
        self.device_scope = kwargs.get("device_scope", None)
        self.parent_scopes = kwargs.get("parent_scopes", None)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Provides authentication classes for use with the msrest library
"""

from msrest.authentication import Authentication, BasicTokenAuthentication
from .connection_string import ConnectionString
from .connection_string import HOST_NAME, SHARED_ACCESS_KEY_NAME, SHARED_ACCESS_KEY
from .sastoken import SasToken
from azure.core.pipeline.policies import BearerTokenCredentialPolicy
from azure.core.pipeline import PipelineRequest, PipelineContext
from azure.core.pipeline.transport import HttpRequest

__all__ = ["ConnectionStringAuthentication", "AzureIdentityCredentialAdapter"]


class ConnectionStringAuthentication(ConnectionString, Authentication):
    """ConnectionString class that can be used with msrest to provide SasToken authentication

    :param connection_string: The connection string to generate SasToken with
    """

    def __init__(self, connection_string):
        super(ConnectionStringAuthentication, self).__init__(
            connection_string
        )  # ConnectionString __init__

    @classmethod
    def create_with_parsed_values(cls, host_name, shared_access_key_name, shared_access_key):
        connection_string = (
            HOST_NAME
            + "="
            + host_name
            + ";"
            + SHARED_ACCESS_KEY_NAME
            + "="
            + shared_access_key_name
            + ";"
            + SHARED_ACCESS_KEY
            + "="
            + shared_access_key
        )
        return cls(connection_string)

    def signed_session(self, session=None):
        """Create requests session with any required auth headers applied.

        If a session object is provided, configure it directly. Otherwise,
        create a new session and return it.

        :param session: The session to configure for authentication
        :type session: requests.Session
        :rtype: requests.Session
        """
        session = super(ConnectionStringAuthentication, self).signed_session(session)

        # Authorization header
        sastoken = SasToken(self[HOST_NAME], self[SHARED_ACCESS_KEY], self[SHARED_ACCESS_KEY_NAME])
        session.headers[self.header] = str(sastoken)
        return session


class AzureIdentityCredentialAdapter(BasicTokenAuthentication):
    def __init__(self, credential, resource_id="https://iothubs.azure.net/.default", **kwargs):
        """Adapt any azure-identity credential to work with SDK that needs azure.common.credentials or msrestazure.
        Default resource is ARM (syntax of endpoint v2)
        :param credential: Any azure-identity credential (DefaultAzureCredential by default)
        :param str resource_id: The scope to use to get the token (default ARM)
        """
        super(AzureIdentityCredentialAdapter, self).__init__(None)
        self._policy = BearerTokenCredentialPolicy(credential, resource_id, **kwargs)

    def _make_request(self):
        return PipelineRequest(
            HttpRequest("AzureIdentityCredentialAdapter", "https://fakeurl"), PipelineContext(None)
        )

    def set_token(self):
        """Ask the azure-core BearerTokenCredentialPolicy policy to get a token.
        Using the policy gives us for free the caching system of azure-core.
        We could make this code simpler by using private method, but by definition
        I can't assure they will be there forever, so mocking a fake call to the policy
        to extract the token, using 100% public API."""
        request = self._make_request()
        self._policy.on_request(request)
        # Read Authorization, and get the second part after Bearer
        token = request.http_request.headers["Authorization"].split(" ", 1)[1]
        self.token = {"access_token": token}

    def signed_session(self, session=None):
        self.set_token()
        return super(AzureIdentityCredentialAdapter, self).signed_session(session)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class DeviceRegistryOperationWarning(Model):
    """The device registry operation warning details.

    :param device_id: The unique identifier of the device.
    :type device_id: str
    :param warning_code: The warning code. Possible values include:
     'DeviceRegisteredWithoutTwin'
    :type warning_code: str or ~protocol.models.enum
    :param warning_status: The details of the warning.
    :type warning_status: str
    """

    _attribute_map = {
        "device_id": {"key": "deviceId", "type": "str"},
        "warning_code": {"key": "warningCode", "type": "str"},
        "warning_status": {"key": "warningStatus", "type": "str"},
    }

    def __init__(
        self, *, device_id: str = None, warning_code=None, warning_status: str = None, **kwargs
    ) -> None:
        super(DeviceRegistryOperationWarning, self).__init__(**kwargs)
        self.device_id = device_id
        self.warning_code = warning_code
        self.warning_status = warning_status



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class Configuration(Model):
    """The configuration for Iot Hub device and module twins.

    :param id: The unique identifier of the configuration.
    :type id: str
    :param schema_version: The schema version of the configuration.
    :type schema_version: str
    :param labels: The key-value pairs used to describe the configuration.
    :type labels: dict[str, str]
    :param content: The content of the configuration.
    :type content: ~protocol.models.ConfigurationContent
    :param target_condition: The query used to define the targeted devices or
     modules. The query is based on twin tags and/or reported properties.
    :type target_condition: str
    :param created_time_utc: The creation date and time of the configuration.
    :type created_time_utc: datetime
    :param last_updated_time_utc: The update date and time of the
     configuration.
    :type last_updated_time_utc: datetime
    :param priority: The priority number assigned to the configuration.
    :type priority: int
    :param system_metrics: The system metrics computed by the IoT Hub that
     cannot be customized.
    :type system_metrics: ~protocol.models.ConfigurationMetrics
    :param metrics: The custom metrics specified by the developer as queries
     against twin reported properties.
    :type metrics: ~protocol.models.ConfigurationMetrics
    :param etag: The ETag of the configuration.
    :type etag: str
    """

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "schema_version": {"key": "schemaVersion", "type": "str"},
        "labels": {"key": "labels", "type": "{str}"},
        "content": {"key": "content", "type": "ConfigurationContent"},
        "target_condition": {"key": "targetCondition", "type": "str"},
        "created_time_utc": {"key": "createdTimeUtc", "type": "iso-8601"},
        "last_updated_time_utc": {"key": "lastUpdatedTimeUtc", "type": "iso-8601"},
        "priority": {"key": "priority", "type": "int"},
        "system_metrics": {"key": "systemMetrics", "type": "ConfigurationMetrics"},
        "metrics": {"key": "metrics", "type": "ConfigurationMetrics"},
        "etag": {"key": "etag", "type": "str"},
    }

    def __init__(self, **kwargs):
        super(Configuration, self).__init__(**kwargs)
        self.id = kwargs.get("id", None)
        self.schema_version = kwargs.get("schema_version", None)
        self.labels = kwargs.get("labels", None)
        self.content = kwargs.get("content", None)
        self.target_condition = kwargs.get("target_condition", None)
        self.created_time_utc = kwargs.get("created_time_utc", None)
        self.last_updated_time_utc = kwargs.get("last_updated_time_utc", None)
        self.priority = kwargs.get("priority", None)
        self.system_metrics = kwargs.get("system_metrics", None)
        self.metrics = kwargs.get("metrics", None)
        self.etag = kwargs.get("etag", None)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class PropertyContainer(Model):
    """The desired and reported properties of the twin. The maximum depth of the
    object is 10.

    :param desired: The collection of desired property key-value pairs. The
     keys are UTF-8 encoded, case-sensitive and up-to 1KB in length. Allowed
     characters exclude UNICODE control characters (segments C0 and C1), '.',
     '$' and space. The desired porperty values are JSON objects, up-to 4KB in
     length.
    :type desired: dict[str, object]
    :param reported: The collection of reported property key-value pairs. The
     keys are UTF-8 encoded, case-sensitive and up-to 1KB in length. Allowed
     characters exclude UNICODE control characters (segments C0 and C1), '.',
     '$' and space. The reported property values are JSON objects, up-to 4KB in
     length.
    :type reported: dict[str, object]
    """

    _attribute_map = {
        "desired": {"key": "desired", "type": "{object}"},
        "reported": {"key": "reported", "type": "{object}"},
    }

    def __init__(self, **kwargs):
        super(PropertyContainer, self).__init__(**kwargs)
        self.desired = kwargs.get("desired", None)
        self.reported = kwargs.get("reported", None)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class DeviceJobStatistics(Model):
    """The job statistics regarding execution status.

    :param device_count: The number of devices targeted by the job.
    :type device_count: int
    :param failed_count: The number of failed jobs.
    :type failed_count: int
    :param succeeded_count: The number of succeeded jobs.
    :type succeeded_count: int
    :param running_count: The number of running jobs.
    :type running_count: int
    :param pending_count: The number of pending (scheduled) jobs.
    :type pending_count: int
    """

    _attribute_map = {
        "device_count": {"key": "deviceCount", "type": "int"},
        "failed_count": {"key": "failedCount", "type": "int"},
        "succeeded_count": {"key": "succeededCount", "type": "int"},
        "running_count": {"key": "runningCount", "type": "int"},
        "pending_count": {"key": "pendingCount", "type": "int"},
    }

    def __init__(self, **kwargs):
        super(DeviceJobStatistics, self).__init__(**kwargs)
        self.device_count = kwargs.get("device_count", None)
        self.failed_count = kwargs.get("failed_count", None)
        self.succeeded_count = kwargs.get("succeeded_count", None)
        self.running_count = kwargs.get("running_count", None)
        self.pending_count = kwargs.get("pending_count", None)



# coding=utf-8
# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ConfigurationQueriesTestInput(Model):
    """ConfigurationQueriesTestInput.

    :param target_condition: The query used to define targeted devices or
     modules. The query is based on twin tags and/or reported properties.
    :type target_condition: str
    :param custom_metric_queries: The key-value pairs with queries and their
     identifier.
    :type custom_metric_queries: dict[str, str]
    """

    _attribute_map = {
        "target_condition": {"key": "targetCondition", "type": "str"},
        "custom_metric_queries": {"key": "customMetricQueries", "type": "{str}"},
    }

    def __init__(
        self, *, target_condition: str = None, custom_metric_queries=None, **kwargs
    ) -> None:
        super(ConfigurationQueriesTestInput, self).__init__(**kwargs)
        self.target_condition = target_condition
        self.custom_metric_queries = custom_metric_queries
